/*
 * pi - P user Interface
 * 
 * $Id: util.c,v 1.1.1.1 2001/02/13 18:30:27 malekith Exp $
 * Author: Michal Moskal <malekith@pld.org.pl>
 * include COPYING-GNU
 */

#include <pi/pi.h>
#include <string.h>
#include <stdlib.h>

void pi_set_color(struct pi_object *ctx, int hl)
{
	if (hl)
		pi_color(pi_is_active(ctx) ? 
			pi_int(color_acct_hl) : pi_int(color_norm_hl));
	else
		pi_color(pi_is_active(ctx) ? 
			pi_int(color_acct) : pi_int(color_norm));
}

void pi_init_size(struct pi_object *ret, int x, int y, int w, int h)
{
	if (ret->x == -1)
		ret->x = x;
	if (ret->y == -1)
		ret->y = y;
	if (ret->w == -1)
		ret->w = w;
	if (ret->h == -1)
		ret->h = h;
}
		
void pi_draw(struct pi_object *ctx)
{
	struct pi_object *p;
	void (*draw)(struct pi_object*);

	draw = pi_ptr(draw_handler);
	if (draw)
		draw(ctx);
	
	for (p = ctx->children; p; p = p->next)
		if (p != ctx->active)
			pi_draw(p);
			
	if (ctx->active)
		pi_draw(ctx->active);
}

void pi_send_cc(struct pi_object *ctx, int child_command)
{
	void (*cc)(struct pi_object*, int);

	ctx = ctx->parent;
	assert(ctx);
	
	cc = pi_ptr(cc_handler);
	if (cc)
		cc(ctx, child_command);
}

void pi_center(struct pi_object *w)
{
	w->x = (pi_screen_width - w->w) / 2;
	w->y = (pi_screen_height - w->h) / 2;
}

void pi_locate(struct pi_object *self, int *xx, int *yy)
{
	int x, y;
	struct pi_object *p;

	x = y = 0;

	for (p = self; p; p = p->parent) {
		x += p->x;
		y += p->y;
	}

	pi_goto(x, y);
	if (xx)
		*xx = x;
	if (yy)
		*yy = y;
}

void pi_setstr(char **p, const char *s)
{
	if (*p == s)
		return;

	free(*p);
	*p = malloc(strlen(s) + 1);
	if (*p == 0)
		pi_die("out of memory");
	strcpy(*p, s);
}

int pi_is_active(struct pi_object *self)
{
	return self->parent == 0 ||
	       (pi_is_active(self->parent) && self->parent->active == self);
}

void pi_setpos(struct pi_object *p, int x, int y, int w, int h)
{
	p->w = w;
	p->h = h;
	p->x = x;
	p->y = y;
}

void pi_scrollbar(int i, int top, int h, int nlines, int color, int color_hl)
{
	if ((i - top < top * h / nlines) ||
	    (i - top > (top + h) * h / nlines)) {
		pi_color(color);
		pi_putch('.');
	} else {
		pi_color(color_hl);
		pi_putch('#');
	}
}
