/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: fileselector.cc,v 1.4 2000/09/09 17:59:14 malekith Exp $ */

#define _V_PRIVATE
#include <v/v.h>
#include <v/term.h>
#include <v/fileselector.h>
#include <v/re.h>

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fnmatch.h>

const int ev_list = ev_user + 10000;

static bool is_dir(String f)
{
	struct stat buf;
	return (stat(f, &buf) == 0 && S_ISDIR(buf.st_mode));
}

void FileListItem::execute(ListBox *)
{
	Event e;
	e.type = Event::command;
	e.data = ev_list;
	e.user_data = user_data;
	send_event(e);
//	p->handle_event(e);
}

FileSelector::FileSelector()
{
	pos.resize_to(50,20);
	center();
	
	iline = new InputLine;
	iline->pos.grow(1, 1, pos.w - 2, 1);
	add_child(iline);
	
	lbox = new ListBox;
	lbox->pos.move_to(1,3);
	lbox->pos.resize_to(pos.w - 2, pos.h - 4);
	add_child(lbox);
	name = "FileSelector";
}

void FileSelector::execute_file(String file)
{
	if (is_dir(file)) {
		String look_for = "";
		if (file *= "^/?/\\.\\./?$")
			return;
		Regex re("^(.*)/([^/]+/)/?\\.\\./?$");
		if (file *= re) {
			wd = re.r(1);
			look_for = re.r(2);
		} else
			wd = file;
		create_listing();
		
		if (look_for != "") {
			int i;
			for (i=0; i<lbox->items.size(); i++)
				if (lbox->items[i]->caption == look_for)
					break;
			if (i<lbox->items.size()) {
				lbox->active_item = i;
				draw();
			}
		}
	} else if (check_file(file))
		dead = true;
}

void FileSelector::new_filter(String f)
{
	
	// do ~ expansion
	Regex re("\\~(.*)");
	if (f *= re)
		f = getenv("HOME") + re.r(1);
	
	// check whatever this is a directory
	if ((f *= "^/") && is_dir(f)) {
		// yap, absolute
		execute_file(f);
		iline->caption = "";
	} else if (is_dir(wd + "/" + f)) {
		// yap, relative
		execute_file(wd + "/" + f);
		iline->caption = "";
	} else {
		// nope, treat as a pattern
		filter = f;
	}
	create_listing();
}

void FileSelector::handle_event(Event & e)
{
	if (e.type == Event::keyboard)
		switch (e.data) {
		case '\n':
			if (active == iline)
				new_filter(iline->caption);
			break;
		// a so-called lynx-like motion ;)
		case key_left:
			if (v_get_int_conf("lynx-like-motion") && 
			    active == lbox) {
				execute_file(wd + "/.."); 
				e.kill();
			}
			break;
		case key_right:
			if (v_get_int_conf("lynx-like-motion") && 
			    active == lbox)
				if (v_get_int_conf("lynx-like-select") || 
				    is_dir(wd + "/" + 
				    lbox->items[lbox->active_item]->caption))
					e.data = '\n';
				else
					e.kill();
			break;
		}

	Window::handle_event(e);

	if (e.type == Event::command && e.data == ev_list)
		execute_file(wd + "/" + 
			     lbox->items[(int) e.user_data]->caption);
}

void FileSelector::create_listing()
{
	DIR *d;
	dirent *e;
	Block <String> dirs, files;
	int k;
	Regex re("(.*)/$");
	
	if (wd *= re)
		wd = re.r(1);
	re.set("^/+(/.*)");
	if (wd *= re)
		wd = re.r(1);
		
	if (wd == "")
		wd = "/";

	lbox->free();
	d = opendir(wd);
	title = wd;
	if (filter == "")
		filter = "*";
	if (iline->caption == "")
		iline->caption = filter;
	
	if (!d)
		throw XWidget("FileSelector::create_listing(): "
			      "couldn't opendir( '" + wd + "' )");
	while ((e = readdir(d))) {
		String name;
		
		if (strcmp(e->d_name,".") == 0)
			continue;
		name.setf("%s/%s", wd.cstr(), e->d_name);
		if (is_dir(name))
			dirs.push(e->d_name);
		else if (fnmatch(filter, e->d_name, FNM_CASEFOLD) == 0)
			files.push(e->d_name);
	}
	closedir(d);
	
	dirs.sort();
	files.sort();
	
	lbox->items.alloc(dirs.size() + files.size());
	
	k = 0;
	for (int i = 0; i < dirs.size(); i++, k++) {
		String s;
		s.setf("%s/", dirs[i].cstr());
		lbox->items[k] = new FileListItem(s, (void*)k);
	}

	for (int i = 0; i < files.size(); i++, k++)
		lbox->items[k] = new FileListItem(files[i], (void*)k);
		
	lbox->top_item = lbox->active_item = 0;
	draw();
}

int FileSelector::execute()
{
	if (wd == "") {
		char cwd[10000];
		wd = "/";
		if (getcwd(cwd, sizeof(cwd)))
			wd = cwd;
	}
	
	file = "";
	create_listing();
	return Dialog::execute();
}
