/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: widgets.cc,v 1.17 2000/09/20 08:24:55 malekith Exp $ */

#define _V_PRIVATE
#include <v/v.h>
#include <v/term.h>

#include <ctype.h>

// ----------------------------------------------------------------------------
// Label
void Label::draw()
{
	int fill_left, fill_right;

	go(pos_x(), pos_y());
	if (!pos.w || !pos.h) 
		compute_sizes();
	
	switch (align) {
	case left:
		fill_left = 0;
		fill_right = pos.w - strlen(caption);
		break;
	case right:
		fill_left = pos.w - strlen(caption);
		fill_right = 0;
		break;
	case center:
		fill_left = (pos.w - strlen(caption) + 1) / 2;
		fill_right = (pos.w - strlen(caption)) / 2;
		break;
	default:
		abort();
	};
	
	set_color(fcolor);
	while (fill_left-- > 0)
		print(' ', fcolor);
	print("%.*s",pos.w,caption.cstr());
	while (fill_right-- > 0)
		print(' ', fcolor);
	go(pos_x(), pos_y());
}

void Label::compute_sizes()
{
	pos.w = strlen(caption);
	pos.h = 1;
}

// ----------------------------------------------------------------------------
// Button
void Button::draw()
{
	int c = is_active() ? color("button_active") : color("button");
	int fill_left, fill_right;

	go(pos_x(), pos_y());
	print('[', c);
	
	if (!pos.w || !pos.h) 
		compute_sizes();
	
	fill_left = (pos.w - 2 - v_alt_strlen(caption) + 1) / 2;
	fill_right = (pos.w - 2 - v_alt_strlen(caption)) / 2;
	
	while (fill_left-- > 0)
		print(' ', c);
	v_alt_print(caption, c,
		    is_active()? color("button_hl_active") :
		    color("button_hl"),
		    pos.w - 2 < v_alt_strlen(caption) ? pos.w - 2 : -1);
	while (fill_right-- > 0)
		print(' ', c);
	print(']', c);
	go(pos_x() + 2, pos_y());
}

void Button::compute_sizes()
{
	pos.w = v_alt_strlen(caption) + 4;
	pos.h = 1;
}

void Button::handle_event(Event & e)
{
	if (
		((e.type == Event::keyboard)
			&& (e.data == ' ' || e.data == '\n')) || 
		((e.type == Event::mouse)
			&& (e.m.event == Event::MouseState::execute))) {
		execute();
		e.kill();
	}
}

bool Button::is_accel(Event &e)
{
	return e.type == Event::keyboard && _v_alt_checkkey(caption, e.data);
}

// ----------------------------------------------------------------------------
// Window
void Window::handle_event(Event & e)
{
	if (e.type == Event::mouse)
		switch (e.m.event) {
		case Event::MouseState::start_drag:
			dragging = false;
			if (e.m.y == pos_y()) {
				dragging = true;
				start_x = e.m.x - pos.x;
				start_y = e.m.y - pos.y;
			}
			e.kill();
			break;
		case Event::MouseState::drag:
			if (!dragging)
				return;
			pos.move_to(e.m.x - start_x, e.m.y - start_y);
			parent->draw();
			e.kill();
			break;
		case Event::MouseState::end_drag:
			if (!dragging)
				return;
			dragging = false;
			e.kill();
			break;
		case Event::MouseState::execute:
			if (close_button && e.m.y == pos_y()
			    && e.m.x == pos_x() + pos.w - 4) {
				dead = true;
				e.kill();
				return;
			}
			break;
		default:
			break;
		}
	else if (e.type == Event::keyboard)
		switch (e.data) {
		case 27:
		case key_ctrl('x'):
			dead = true;
			e.kill();
			break;
		}
	Form::handle_event(e);
}

void Window::draw()
{
	int fc = is_active() ? color("win_frame_active") : color("win_frame");
	int tc = is_active() ? color("win_title_active") : color("win_title");
	
	// draw a frame
	pos.draw_frame_at(pos_x(), pos_y(), fc, true);
		
	// and a title
	if (pos.w > 12) {
		String t; 
		if (title.strlen() > pos.w - 10)
			t.setf("%.*s...", pos.w - 12, title.cstr());
		else
			t = title;
		go(pos_x() + (pos.w - t.strlen()) / 2 - 1, pos_y());
		set_color(tc);
		print(" %s ", t.cstr());
	}
	
	// close button
	go(pos_x() + pos.w - 5, pos_y());
	print('[',fc); print(close_button ? 'X' : 'O',tc); print(']',fc);
	
	// and let children draw themself
	Form::draw();
}

void Window::center()
{
	pos.move_to((s_width - pos.w) / 2, (s_height - pos.h) / 2);
}

// ----------------------------------------------------------------------------
// RootWindow
void RootWindow::draw()
{
	set_color(color("background"));
	cls();
	Form::draw();
}

void RootWindow::handle_event(Event &e)
{
	if (e.type == Event::keyboard && 
	    e.data & key_meta && menu && !antimenu &&
	    menu->is_accel(e)) {
		send_command(ev_menu_start);
		e.kill();
		return;
	}
	
	Form::handle_event(e);
	
	switch (e.type) {
	case Event::keyboard:
		handle_key(e.data);
		e.kill();
		break;
	case Event::command:
		switch (e.data) {
		case ev_quit:
			dead = true;
			break;
		case ev_repaint:
			draw();
			break;
		case ev_menu_start:
			if (menu && active != menu) {
				antimenu = active;
				active = menu;
				active->draw();
			}
			break;
		case ev_menu_stop:
			if (antimenu) {
				active = antimenu;
				antimenu = 0;
				draw();
			}
			break;
		default:
			v_handle_command(e.data);
			break;
		}
		break;
	case Event::mouse:
	case Event::dead:
		return;
	default:
		throw XWidget("RootWindow::handle_event(): unknown event type");
	}
}

void RootWindow::handle_key(int k)
{
	switch (k) {
	case key_ctrl('L'):
		draw();
		f_repaint();
		break;
	case key_ctrl('N'):
		focus_next();
		draw();
		break;
	case key_meta | 'q':
	case key_meta | 'Q':
	case key_f(10):
	case key_ctrl('C'):
		send_command(ev_quit);
		dead = true;
		break;
	case key_f(9):
		send_command(ev_menu_start);
		break;
	}
}

// ----------------------------------------------------------------------------
// Dialog
void Dialog::handle_event(Event &e)
{
	Window::handle_event(e);
		
	if (e.type == Event::command && e.data >= ev_ok) {
		modal_result = e.data;
		dead = true;
		e.kill();
	} else if ((e.type == Event::keyboard && e.data == 27) || 
		   (e.type == Event::command && e.data == ev_quit))  {
		modal_result = 0;
		dead = true;
		e.kill();
	} 
}

int Dialog::execute()
{
	modal_result = 0;
	event_loop();
	return modal_result;
}

// ----------------------------------------------------------------------------
// TextArea
int TextArea::word_break(int width, String &result)
{
	String buf(0, width + 10), r;
	int last = -1, minus = 1;
	const char *s = caption;
	char *d = buf.cstr();
	int lines = 0;
	
	result = "";
	
	for (int n = 0; *s; s++, n++) {
		if (*s == '\n') {
			*d++ = '\n'; *d = 0;
			goto add_line;
		} else if (isspace(*s))
			last = n;
		
		*d++ = *s;
			
		if (n >= width) {
			if (last == -1) {
				minus = -1;
				*d++ = '\n'; *d = 0;
			} else {
				if (n == last) {
					while (*s == ' ' || *s == '\t')
						s++;
					if (*s == 0)
						goto bye;
					s--;
				} else
					s -= n - last;
				buf.cstr()[last] = '\n';
				buf.cstr()[last + 1] = 0;
			}
		} else
			continue;
		
	add_line:
		result += buf;
		d = buf.cstr();
		n = last = -1;
		lines++;
	}

bye:
	*d = 0;
	result += buf;
	lines++;
	return minus * lines;
}

void TextArea::draw()
{
	int x = pos_x(), y = pos_y();
	String to_draw;
	
	if (!pos.w)
		pos.w = 60;

	pos.h = word_break(pos.w, to_draw);
	set_color(fcolor);

	for (char *p = to_draw.cstr(); p && *p;) {
		char *np;
		int fill_left=0, fill_right=0;
		
		if ((np = strchr(p, '\n')))
			*np++ = 0;
			
		go(x, y++);
		
		switch (align) {
		case left:
			fill_left = 0;
			fill_right = pos.w - strlen(p);
			break;
		case right:
			fill_left = pos.w - strlen(p);
			fill_right = 0;
			break;
		case center:
			fill_left = (pos.w - strlen(p) + 1) / 2;
			fill_right = (pos.w - strlen(p)) / 2;
			break;
		};

		while (fill_left-- > 0)
			print(' ', fcolor);
		print("%.*s", pos.w, p);
		while (fill_right-- > 0)
			print(' ', fcolor);

		p = np;
	}

	go(pos_x(), pos_y());
}

// ----------------------------------------------------------------------------
// StatusLine
void StatusLine::draw()
{
	Widget *w = v_root_window.active;

	if (w == 0)
		return;

	while (w->active)
		w = w->active;

	while (w && w->help == "")
		w = w->parent;

	if (w == 0)
		return;

	go(0, s_height - 1);
	v_alt_print(w->help, color("status_normal"), color("status_hl"),
		    s_width);

	if (right_side != "") {
		go(s_width - v_alt_strlen(right_side), s_height - 1);
		v_alt_print(right_side, color("status_normal"),
			    color("status_hl"));
	}
	if (down != "") {
		go(0, s_height - 2);
		v_alt_print(down, COLOR(LIGHTGRAY, BLACK), COLOR(WHITE, BLACK));
	}
}

// ----------------------------------------------------------------------------
// ProgressWin
ProgressWin::ProgressWin(int max, const String &t, const String &i)
{
	pos.resize_to(50, 5);
	center();
	
	add_child(info = new Label(2, 1, i, 0, Label::center, 46));
	add_child(pb = new ProgressBar(2, 3, 46, 0, max));
	title = t;
	pb->frame = 0;
}

void ProgressWin::set(int p, int max)
{
	if (max)
		pb->max = max;
	pb->advance = p;
	
	if (pb->changed()) {
		draw();
		repaint();
	}
}

// ----------------------------------------------------------------------------
// ProgressBar
void ProgressBar::count()
{
	int p, m, w;

	w = pos.w - (frame ? 2 : 0);
	if (w < 1)
		w = 1;

	p = advance < 0 ? -advance : advance;
	m = max < 0 ? -max : max;
	if (m == 0)
		m = 100;
	if (p > m)
		p = m;
	
	filled = (int)(w * ((double)p / m));
	unfilled = w - filled;
}

bool ProgressBar::changed()
{
	count();
	return (filled != sv_filled || unfilled != sv_unfilled);
}

void ProgressBar::draw()
{
	int c1 = color("pbar_frame");
	int cf = color("pbar_fill");
	int ce = color("pbar_empty");
	
	count();
	
	sv_filled = filled;
	sv_unfilled = unfilled;
	
	go(pos_x(), pos_y());
	
	if (frame)
		print('[', c1);
	while (filled--)
		print('#', cf);
	while (unfilled--)
		print(' ', ce);
	if (frame)
		print(']', c1);
}

// ----------------------------------------------------------------------------
// Box
void Box::draw()
{
	int fc = color("win_frame");
	int tc = color("win_title");

	// draw a frame
	pos.draw_frame_at(pos_x(), pos_y(), fc, true);
		
	// and a title
	if (pos.w > 12) {
		String t; 
		if (title.strlen() > pos.w - 10)
			t.setf("%.*s...", pos.w - 12, title.cstr());
		else
			t = title;
		go(pos_x() + (pos.w - t.strlen()) / 2 - 1, pos_y());
		set_color(tc);
		print(" %s ", t.cstr());
	}
}

